options(stringsAsFactors = FALSE, scipen=3)

rm(list=ls())
gc(verbose=FALSE)

# --------------------------------------------------------------------------------------
# MSE of Faroe Plateau cod        requires R vers 2.8.1
# Author  : Clara Ulrich - with someinputs from Jan Jaap Poos's assessment script
# used for the determination of Fmsy given a Btrigger (Bpa) at the benchmark (WKFLAT) and further at WGBFAS2011
# 29/04/2011  - modified by Jesper Boje
# --------------------------------------------------------------------------------------
assyear <- 2010
### ------------------------------------------------------------------------------------------------------
###  1. libraries
### ------------------------------------------------------------------------------------------------------
library(FLXSA); library(FLash);#library(FLBRP)
library(FLAdvice)
library(abind)

# FLXSA calls FLCore and FLAssess directly


### ------------------------------------------------------------------------------------------------------
###  2. set paths
### ------------------------------------------------------------------------------------------------------
stock.name   <- "fsai"
#drive        <- "D:"
drive        <- ""
#my.path      <- "\\Stovnsmeting2011\\fsai\\fsai\\"
#my.path      <- "/media/Data/Stovnsmeting2011/fsai/fsai/"        # Run on Ubuntu 11.04 and R 2.12.0
my.path      <- "/home/luridao/Documents/D/Stovnsmeting2012/fsai/sim/"	# Run on Ubuntu 10.04.04 and R 2.14.2
#my.path      <- "/windows/D/Stovnsmeting2011/fsai/fsai/"         # Run on openSUSE 11.4 (i586) and R 2.13.1
stock.path   <- paste(drive,my.path, sep="")
result.path  <- paste(drive,my.path, sep="")

  
source(paste(result.path,"funcs.r",sep=""))
### totalStk
## Add up totals for stock objects, computes landings, discards, catch, stock totals
totalStk <- function(stk){
    landings(stk) <- computeLandings(stk)
    discards(stk) <- computeDiscards(stk)
    catch.n(stk)  <- landings.n(stk)+discards.n(stk)
    catch.wt(stk) <-(landings.n(stk)*landings.wt(stk)+discards.n(stk)*discards.wt(stk))/catch.n(stk)
    catch(stk)    <- computeCatch(stk)
    stock(stk)    <- computeStock(stk)
    units(stk)[1:17] <- as.list(c(rep(c("tonnes","thousands","kg"),4), "NA", "NA", "f", "NA", "NA"))
    return(stk)
    }


### ------------------------------------------------------------------------------------------------------
###   3. Read and process assessment input data
### ------------------------------------------------------------------------------------------------------

stock              <- readFLStock(paste(stock.path,"SAI_IND", sep=""), no.discards=T)
#clara
#landings.wt(stock)[1,] <- landings.wt(stock)[2,]
#catch.wt(stock)[1,] <- catch.wt(stock)[2,]
#stock.wt(stock)[1,] <- stock.wt(stock)[2,]
#landings.n(stock)[1,] <- 1
#catch(stock)       <- computeCatch(stock,slot="all")
stock              <- setPlusGroup(stock,plusgroup=14)
#units(stock)[1:17] <- as.list(c(rep(c("tonnes","thousands","kg"),4), "NA", "NA", "f", "NA", "NA"))
indices            <- readFLIndices(paste(stock.path,"pt_glmSD_adj2.dat", sep=""), na.strings="-1")

# set correct data in range
range(stock)[["minfbar"]] <- 4
range(stock)[["maxfbar"]] <- 8
#units(harvest(stock))     <- "f"

# do sop correction
#soplan             <- sop(stock,"landings")
#range(soplan)
#stock@landings.wt  <- sweep(stock@landings.wt,2,soplan,"/")
#sopc               <- sop(stock,"landings")

#NOTE FROM CLARA : IF YOU DO THE SOP CORRECTION YOU SHOULD UPDATE THE CATCH.WT AS WELL!
#clara
# Add up totals
stock <- totalStk(stock)


### ------------------------------------------------------------------------------------------------------
###   4. Single XSA assessment, SR
### ------------------------------------------------------------------------------------------------------


xsa.indices  <- indices#FLIndices( trim(indices[[1]], year=1996:2010, age=2:8),
                #           trim(indices[[2]], year=1993:2010, age=1:8)
                #                                        )



# control object
xsa.control <- FLXSA.control(tol = 1e-09
, maxit = 30
, min.nse = 0.3
, fse = 2
, rage = 0
, qage = 8
, shk.n = TRUE, shk.f = TRUE
, shk.yrs = 5, shk.ages = 3
, tsrange = 100, tspower = 0
, vpa = FALSE)

names(xsa.indices) <-  lapply(xsa.indices, name)
xsa.res     <- FLXSA(stock, xsa.indices, xsa.control)
names(xsa.res@index.res) <-  lapply(xsa.indices, name)

#check convergence (maxit inf toxsa.control input)
xsa.res@control@maxit

stock   <- transform(stock, harvest=xsa.res@harvest, stock.n=xsa.res@stock.n)

#average logq res. cv of indices
logq.sd <- sapply(xsa.res@index.res, function(x) {
                sd.q <- sd(matrix(x,dim(x)[2],dim(x)[1],byrow=T),na.rm=T)
                sd.q <-mean(sd.q[sd.q<1]) })
logq.sd

#stock <- window(stock, end=assyear)

## Stock recruitment relationship
fsaiSR       <-as.FLSR(stock)
model(fsaiSR)<-ricker()
fsaiSR       <-transform(fsaiSR,ssb=ssb/1000,rec=rec/1000)
fsaiSR       <-fmle(fsaiSR)
plot(fsaiSR)
#savePlot("fsai_SR.png", type="png")
params(fsaiSR)["b",1]<-params(fsaiSR)["b",1]/1000
sd. <- sd(residuals(fsaiSR))


### ------------------------------------------------------------------------------------------------------
###   5. CONDITIONING OF THE MSE
### ------------------------------------------------------------------------------------------------------

## set up years
nits                   <-1000  # number of iterations
now                    <-assyear+1   # current year, i.e. starting year for MSE
history.yr             <-dims(stock)$minyear
proj.yr                <- 40
future.yrs             <-now:(now+proj.yr-1) # final project year
plusgroup              <-stock@range["plusgroup"]
yrs.assess             <-ac(history.yr:(now-1))

OEM.XSA <- FALSE

## set up objects

OM                   <-propagate(stf(stock,nyears=(proj.yr+1)),iter=nits)	# propagate(stf(stock,nyears=(proj.yr+1)),iter=nits)
ages                 <- stock@range["min"]:stock@range["max"]
dmns                 <-list(age=ages,year=history.yr:(now+proj.yr),iter=1:nits)

MP                   <-OEMCatch(OM,start=history.yr,end=now-1)

#CHOICE OF AUTOCORR MODEL
autocor. <- FALSE

if (autocor. ==F) {sd.sim <- sd.
                    acf. <- 0
} else            {sd.sim <-  sqrt(var(x)*(1-ACF.^2))
                  acf. <- ACF.
                  }


#usual SR
sr.deviates       <-FLQuant(exp(rnorm(nits*(length(future.yrs)+1), 0,sd.)),dimnames=list(age=1,year=now:(now+proj.yr),iter=1:nits))/exp(sd.^2/2.0)

# SR with autocorrelation  - NOT exponential
sr.deviates       <-FLQuant(rnorm(nits*length(future.yrs)+1, 0,sd.sim),dimnames=list(age=2,year=now:(now+proj.yr),iter=1:nits))- sd.sim^2/2.0
sr.resid <- sr.deviates

# initialisation OM 2011

TAC       <-20000#TAC in 2011

ctrl.now <- fwdControl(data.frame(year=now,val=1,rel=now, quantity="f"))
sr.resid[,ac(now)] <- exp(sr.deviates[,ac(now)])
                                            
                                            
OM <-fwd(OM,ctrl=ctrl.now,sr=fsaiSR,sr.residuals=sr.resid) # projection of the survivors at start of 2009


## set up CPUE index
index                               <-as(OM,"FLIndex") # this bugs since it retuns catch.n instead of stock.n, must be checked before use
index@range[c("startf","endf")]     <-c(0,0.1)
index@range <- index@range[!names(index@range) %in% c("minfbar","maxfbar")]

index.deviates    <-FLQuant(exp(rnorm(prod(unlist(lapply(dmns,length))),0,min(logq.sd))),dimnames=dmns)
index@index[,yrs.assess]<-OEMIndex(OM,index.deviates,start=history.yr,end=now-1,plusgroup=plusgroup)



### ------------------------------------------------------------------------------------------------------
###   5. HARVEST CONTROL RULE
### ------------------------------------------------------------------------------------------------------
Bpa          <-60000
Ftarg          <-0.32
TAC.bound    <-1.00

#Or get the BRPs :
fsaiBrp <- FLBRP(stock,sr=fsaiSR)
fsaiBrp <- brp(fsaiBrp)
plot(fsaiBrp,obs=TRUE)
Fmsy <- as.numeric(refpts(fsaiBrp)["msy","harvest"])
Fmsy

## ----------------------- WORKING ON IT AT THE MOMENT -------------------------

cvcW                  <- catch.wt(OM)
cvcW[,ac(1961:2010)]  <- 0
cvcW[,ac(2011:2051)]  <- NA
cvsW                  <- cvcW	# catch weight = stock weight

wgtRho     <- 0.35#0.35
wgtCV      <- 0.2#0.2
wgtErrType <- 2	# 1=Normal   2=logNormal

x <- array(rnorm(proj.yr*nits),
            dim=c(proj.yr,nits),
            dimnames=list(year=now:(now+proj.yr-1),
                          iter=1:nits))
for (i in 2:proj.yr) x[i,] <- wgtRho*x[i-1,]+sqrt(1-wgtRho^2)*x[i,]
x <- wgtCV*x
for (j in ac(ages)) {
#for (i in 1:nits) 
cvcW[j,ac(2011:2050),,,,] <- x	
}

cvsW <- cvcW  # assume same error structure in stock weights as in catch weigts

if(wgtErrType==1){	# 1=Normal   2=logNormal
	OM@catch.wt <- OM@catch.wt * (1+cvcW)
}
if(wgtErrType==2){
	OM@catch.wt <- OM@catch.wt * exp(cvcW)
}

# OM catch weight = OM stock weight
OM@stock.wt <- OM@catch.wt

#- requires library "abind"
#- mean catch-weight
mwtmp	<-	apply(OM@catch.wt, c(1,2),mean,na.rm=T)[,,drop=T]
sdwtmp	<-	apply(OM@catch.wt, c(1,2),sd,na.rm=T)[,,drop=T]
#- mean catch-weight in arrary form
mwtmp	<-	array(mwtmp,dim=c(1,12,91))
#- catch-weight quantiles
qwtmp	<-	apply(OM@catch.wt, c(1,2),function(x) quantile(x,c(0.025, 0.05, 0.25,0.5, 0.75, 0.95, 0.975),na.rm=T))
#- combine mean & quantile catch-weights
qmw		<-	abind(qwtmp, mwtmp, sdwtmp, along=1)
dimnames(qmw)[[1]][8:9] <- c('mean', 'sd')


# ------------------------- ((((((1)))))) --------------------------------------
# ------ intermediate objects to construct the 'tmp's data frame for projections ------- #
ry   <- rep(2011:2049,each=2)          	# rel.year
for(i in seq(1,78,by=2)) ry[i]<-NA 	# rel.year
min. <- rep(1.0,78)          			# min
for(i in seq(1,78,by=2)) min.[i]<-NA	# min
max. <- rep(1.0,78)          			# max
for(i in seq(1,78,by=2)) max.[i]<-NA	# max

# - temporary data frame to construct the 'fwdControl' object
tmp      <- data.frame(year=rep(2011:2050,each=2)
							, val=c(0.32,40000,rep(c(0.32,NA),39))
							, min=c(NA,NA,min.)
							, max=c(NA,NA,max.)
							, quantity=rep(c('f','catch'),40)
							, rel.year=c(NA,NA,ry))
# ------------------------- ((((((2)))))) --------------------------------------
# ------ intermediate objects to construct the 'tmp's data frame for projections ------- #
ry   <- rep(2011:2049,each=2)          	# rel.year
for(i in seq(1,78,by=2)) ry[i]<-NA 	# rel.year
min. <- rep(0.85,78)          			# min
for(i in seq(1,78,by=2)) min.[i]<-NA	# min
max. <- rep(1.15,78)          			# max
for(i in seq(1,78,by=2)) max.[i]<-NA	# max

tmp      <- data.frame(year=rep(2011:2050,each=2)
							, val=c(0.32,80000,rep(c(0.32,NA),39))
							, min=c(NA,NA,min.)
							, max=c(NA,NA,max.)
							, quantity=rep(c('f','ssb'),40)
							, rel.year=c(NA,NA,ry))

# ------------------------- ((((((3)))))) --------------------------------------
# ------ intermediate objects to construct the 'tmp's data frame for projections ------- #
ry   <- rep(2011:2049,each=2)          	# rel.year
for(i in seq(1,78,by=2)) ry[i]<-NA 	# rel.year
min. <- rep(0.8,78)          			# min
for(i in seq(1,78,by=2)) min.[i]<-NA	# min
max. <- rep(1.2,78)          			# max
for(i in seq(1,78,by=2)) max.[i]<-NA	# max

tmp      <- data.frame(year=rep(2011:2050,each=2)
							, val=c(0.32,35000,rep(c(0.32,NA),39))
							, min=c(NA,NA,min.)
							, max=c(NA,NA,max.)
							, quantity=rep(c('f','catch'),40)
							, rel.year=c(NA,NA,ry))

### ------------------------------------------------------------------------------------------------------
###   5. RUNNING THE MSE
### ------------------------------------------------------------------------------------------------------
t1 <- Sys.time()
  cat(sprintf("\n===========================================\n"))
  for (iyr in future.yrs)
     {
     t2 <- Sys.time()
    cat(sprintf("\n===========================================\n"))

     MP@range["maxyear"] <-iyr-1
     yrs.assess           <-as.character((history.yr):MP@range["maxyear"])
     cat(sprintf("Year =%.0f\n",iyr))

     ## No Observation Error Model on the catch
     MP[ ,ac(iyr-1)]<-OEMCatch(OM, start=iyr-1,plusgroup=plusgroup, oem.deviates=NULL)

     #cat("=============","Assess!"," =============\n")
     if (OEM.XSA==TRUE) { #perform an XSA
     index@index[,ac(iyr-1)]<-OEMIndex(OM,deviates=index.deviates,iyr-1,plusgroup=plusgroup)

     MP<-MP[,yrs.assess]+FLXSA(MP[,yrs.assess],trim(index,year=((history.yr):(iyr-1))),
                                  control=xsa.control, diag.flag=FALSE)

     } else { # Perfect assessment : MP = OM, with no deviates (or a random error)
        MP[ ,ac(iyr-1)]<-PerfAssm.st(OM, start=iyr-1,plusgroup=plusgroup, oem.deviates=NULL)
      }



     ## HCR
     #cat("=============","STF and HCR"," =================\n")
     ## years for projection etc.
     yr.now     <-as.character(MP@range["maxyear"]+1)
     yr.TAC     <-as.character(MP@range["maxyear"]+2)
     yr.TACp1     <-as.character(MP@range["maxyear"]+3)
     yr.recruits<-as.character((MP@range["minyear"]+1):(MP@range["maxyear"]-2))

     ## set up future years
     MP   <-stf(window(MP,end=MP@range["maxyear"]),nyears=2)
     ## F status quo
     Fsq   <-c(fbar(MP)[,yr.now])

     # SR residuals on geometric mean
    mean.rec <- apply(rec(MP)[,yr.recruits],6,function(x) exp(mean(log(x))))
    SRrs <- FLQuant(dimnames=list(year=0:2+iyr,iter=1:nits))
    for (i in dimnames(SRrs)$year) SRrs[,i] <- mean.rec
       #HCR Target

      target <-fwdControl(data.frame(year=c(0,1,1,1)+iyr,
       val=c(mean(Fsq),mean(pmin(Fsq,Ftarg)),NA,2),
       min=c(NA,NA,mean(TAC)*(1.0-TAC.bound),NA),
       max=c(NA,NA,mean(TAC)*(1.0+TAC.bound),NA),
       quantity=c(rep("f",2),"catch","f")))


      ## targets by iteration
      dmns      <-dimnames(target@trgtArray)
      dmns$iters<-1:nits

      target@trgtArray<-array(NA,dim=unlist(lapply(dmns,length)),dimnames=dmns)
      target@trgtArray[1,"val",,]<-c(Fsq)
      target@trgtArray[2,"val",,]<-Ftarg   # pmin(Fsq,Ftarg)

      target@trgtArray[3,"min",,]<-c(TAC*(1.0-TAC.bound))
      target@trgtArray[3,"max",,]<-c(TAC*(1.0+TAC.bound))
      target@trgtArray[4,"max",,]<-2
	print("OK to here1")
      ## project MP to TAC year with the HCR and geom mean recr.
      MP<-fwd(MP,ctrl=target,sr=list(model="mean",params=FLPar(1)),sr.residuals=SRrs)	## HERE IS THE PROBLEM
	print("OK to here2")

      ## get Quota
      TAC<-computeCatch(MP)[,yr.TAC]
      cat ("=== mean TAC",mean(TAC),"\n")

      ## go fish!!
      #cat("=============","Go Fish!"," ============\n")
      #project the OM with the TAC constraint and real SR with random error
	print("OK to here3")
     ctrl<-fwdControl(data.frame(year=yr.TAC,max=c(NA,2),val=c(mean(TAC),NA), quantity=c("catch","f")))

# and by iteration
    dms<-dimnames(ctrl@trgtArray)
     dms$iters<-1:nits
     ctrl@trgtArray<-array(NA,lapply(dms,length),dms)
     ctrl@trgtArray[1,"val",,]<-TAC
     ctrl@trgtArray[2,"max",,]<-2.0

     sr.resid[,ac(iyr+1)] <- exp(acf.*sr.deviates[,ac(iyr)]+sr.deviates[,ac(iyr+1)])
     OM <-fwd(OM,ctrl=ctrl,sr=fsaiSR,sr.residuals=sr.resid)

     cat ("time ",Sys.time()-t2,"\n")

    # print(plot(lapply(FLStocks(OM,MP),window,start=1999, end=(iyr+1))))

      } #end of year loop
#cat("tot time",Sys.time()-t1,"\n")

#cat ("===============RESULTS=========\n")

#yearframe <- 2018:(now+proj.yr-1)
#cat ("=====P<Bpa,", length(which(c(ssb(OM)[,ac(yearframe)])<Bpa))/length(c(ssb(OM)[,ac(yearframe)])),"\n")
#cat ("=====mean catch,", mean(computeCatch(OM)[,ac(yearframe)]),"\n")
# plot(ssb(OM)[,as.character(2013:2030)])

    #  save(OM,MP,file=paste(result.path,"results_F0_05TAC10.Rdata",sep=""))

    # savePlot(filename=paste(result.path,"results_F0_05TAC10.wmf",sep=""), type=c("wmf"),device=dev.cur())

#plot(rec(OM))
#plot(fbar(OM))
# plot(OM)

